/*
 * net/sched/sch_dup.c	Duplicate packet scheduler routines.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Catalin(ux aka Dino) BOIE, <catab at deuroconsult.ro>
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/capability.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/tcp.h>
#include <linux/time.h>

#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#include <linux/string.h>
#include <linux/netdevice.h>
#include <net/pkt_sched.h>

#define MODULE_NAME "dup v0.1"

/* global variables */

/* qdisc internal data */
struct dup_sched_data {
	__u32 gap;		/* gap + 1 between duplicate packets */
				/* 0 = disable */
				/* 1 = dup every packet */
				/* 2 = 1 normal, one duplicated */
	__u32 limit;		/* in packets */
	__u32 counter;
};

/* qdisc options */
struct tc_dup_qopt {
	__u32 gap;
	__u32 limit;
};

static int dup_init(struct Qdisc *sch, struct rtattr *opt)
{
	struct dup_sched_data *q = (struct dup_sched_data *)sch->data;

	q->counter = 0;

	if (!opt) {
		q->limit = sch->dev->tx_queue_len;
		q->gap = 0;
	} else {
		struct tc_dup_qopt *ctl = RTA_DATA(opt);

		if (opt->rta_len < RTA_LENGTH(sizeof(*ctl)))
			return -EINVAL;

		q->limit = ctl->limit;
		q->gap = ctl->gap;
	}

	if (q->limit == 0)
		q->limit = 100;

	return 0;
}

static int dup_enqueue(struct sk_buff *skb, struct Qdisc *sch)
{
	struct dup_sched_data *q = (struct dup_sched_data *)sch->data;

	/* increment counter */
	q->counter++;

	printk(KERN_DEBUG "%s: enqueue:"
	       " len=%d Q%X:%X gap=%d counter=%d\n",
	       MODULE_NAME, skb->len,
	       sch->handle >> 16, sch->handle & 0xffff, q->gap, q->counter);

	/* normal send */
	if (sch->q.qlen < q->limit) {
		struct sk_buff *clone;

		__skb_queue_tail(&sch->q, skb);
		sch->stats.bytes += skb->len;
		sch->stats.packets++;

		/* clone skb */
		clone = skb_clone(skb, GFP_ATOMIC);
		if (clone) {
			/* Ignore queue size. We really need to send a duplicate. :) */
			if ((q->gap > 0) && (q->counter >= q->gap)) {
				printk(KERN_DEBUG "%s: duplicate!\n",
				       MODULE_NAME);

				/* reset counter */
				q->counter = 0;

				/* send duplicate */
				__skb_queue_tail(&sch->q, clone);
				sch->stats.bytes += clone->len;
				sch->stats.packets++;
			} else {
				kfree_skb(clone);
			}
		}

		return 0;
	}

	sch->stats.drops++;
	kfree_skb(skb);

	return NET_XMIT_DROP;
}

static int dup_requeue(struct sk_buff *skb, struct Qdisc *sch)
{

	__skb_queue_head(&sch->q, skb);

	return 0;
}

static struct sk_buff *dup_dequeue(struct Qdisc *sch)
{
	struct sk_buff *skb;

	skb = __skb_dequeue(&sch->q);

	return skb;
}

static unsigned int dup_drop(struct Qdisc *sch)
{
	struct sk_buff *skb;

	skb = __skb_dequeue_tail(&sch->q);
	if (skb) {
		unsigned int len = skb->len;

		sch->stats.backlog -= len;
		kfree_skb(skb);
		return len;
	}

	return 0;
}

static void dup_reset(struct Qdisc *sch)
{

	skb_queue_purge(&sch->q);
	sch->stats.backlog = 0;
}

static int dup_dump(struct Qdisc *sch, struct sk_buff *skb)
{
	struct dup_sched_data *q = (struct dup_sched_data *)sch->data;
	struct tc_dup_qopt opt;
	unsigned char *b = skb->tail;

	opt.limit = q->limit;
	opt.gap = q->gap;
	RTA_PUT(skb, TCA_OPTIONS, sizeof(opt), &opt);

	return skb->len;

      rtattr_failure:
	skb_trim(skb, b - skb->data);
	return -1;
}

static struct Qdisc_ops dup_qdisc_ops = {
	.next = NULL,
	.cl_ops = NULL,
	.id = "dup",
	.priv_size = sizeof(struct dup_sched_data),
	.enqueue = dup_enqueue,
	.dequeue = dup_dequeue,
	.requeue = dup_requeue,
	.drop = dup_drop,
	.init = dup_init,
	.reset = dup_reset,
	.destroy = NULL,
	.change = dup_init,
	.dump = dup_dump,
	.owner = THIS_MODULE,
};

static int __init init_dup(void)
{
	int ret;

	printk(KERN_DEBUG "%s: (C)opyright Catalin(ux) BOIE 2002-2003\n",
	       MODULE_NAME);

	ret = register_qdisc(&dup_qdisc_ops);
	if (ret != 0) {
		printk(KERN_DEBUG "%s: cannot register qdisc dup. Sorry!\n",
		       MODULE_NAME);
		return -ENOMEM;
	}

	return 0;
}

static void __exit exit_dup(void)
{
	int ret;

	printk(KERN_DEBUG "%s: Goodbye!\n", MODULE_NAME);

	ret = unregister_qdisc(&dup_qdisc_ops);
	if (ret != 0) {
		printk(KERN_DEBUG "%s: Cannot unregister qdisc dup. Sorry!\n",
		       MODULE_NAME);
	}
}

module_init(init_dup);
module_exit(exit_dup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Catalin(ux) BOIE - <catab at deuroconsult.ro>");
MODULE_DESCRIPTION("sch_dup - Duplicate every n packet");
